from typing import Any, Dict, List, Tuple, Union

from lupin_grognard.core.commit.commit import Commit, classify_commits_by_version
from lupin_grognard.core.config import JAMA_TAG
from lupin_grognard.core.doc_generator.jinja_generator import JinjaGenerator
from lupin_grognard.core.git import Git
from lupin_grognard.core.tools.log_utils import info
from lupin_grognard.core.tools.utils import get_current_branch_name


class Changelog(JinjaGenerator):
    def __init__(self, commits: List[Commit]):
        self.commits = commits
        self.git = Git()

    def generate(self, path: str) -> None:
        """Generate changelog"""
        self.git.ensure_not_shallow_clone()

        project_details = self._get_project_details()
        classified_commits = self._classify_commits()
        self._generate_file(
            path=path,
            file_name="changelog.md",
            context={
                "version_details": classified_commits,
                "project_details": project_details,
            },
        )

    def _generate_file(self, path: str, file_name: str, context: Dict) -> None:
        return super()._generate_file(path, file_name, context)

    def _get_project_details(self) -> Dict[str, Union[str, int]]:
        project_url = self.git.get_remote_origin_url()
        project_name = project_url.split("/")[-1]
        info(msg=f"Collecting data from {project_name}")
        branch_name = get_current_branch_name()
        commit_count = self._count_commits()
        first_commit_date = self.git.get_first_commit_date()
        last_commit_date = self.git.get_last_commit_date()
        return {
            "name": project_name,
            "url": project_url,
            "branch_name": branch_name,
            "commit_count": commit_count,
            "first_commit_date": first_commit_date,
            "last_commit_date": last_commit_date,
        }

    def _count_commits(self) -> int:
        number_of_commits = 0
        for commit in self.commits:
            if not commit.is_gitlab_merge_commit():
                number_of_commits += 1
        return number_of_commits

    def _classify_commits(self) -> List[Dict[str, Any]]:
        """
        Classify commits by version and by type and scope
        Returns:
            ListList[Dict[str, Any]]: List of version with commits classified by type and scope

            Example:
            [
                {
                    "version": "v1.0.0",
                    "date": "2020-02-20",
                    "commits": {
                        "feature": {
                            "added": [
                                {
                                    "title": "Add a new feature",
                                    "description": ["line 1", "line 2"],
                                    "gitlab_issue_id": 1,
                                    "gitlab_issue_url": "https://gitlab.com/-/issues/1",
                                }
                            ],
                            "changed": [
                                {
                                    "title: "Change a new feature",
                                    "description": None,
                                    "gitlab_issue_id": 2,
                                    "gitlab_issue_url": "https://gitlab.com/-/issues/2",
                                }
                            ],
                            "removed": [
                                {
                                    "title: "Remove a new feature",
                                    "description": None,
                                }
                            ],
                        },
                        "fixbug": [
                            {
                                "title: "Fix a bug",
                                "description": None,
                            }
                        ],
                        "other": [
                            {
                                "title: "Other commit",
                                "description": None,
                                "gitlab_issue_id": 3,
                                "gitlab_issue_url": "https://gitlab.com/-/issues/3",
                            }
                        ],
                        "unspecified": [
                            {
                                "title: "Unspecified commit",
                                "description": None,
                                "gitlab_issue_id": 4,
                                "gitlab_issue_url": "https://gitlab.com/-/issues/4",
                            }
                        ],
                    },
                },
            ]
        """
        versioned_commits = classify_commits_by_version(self.commits, self.git)
        for v in versioned_commits:
            classified_commits = self._classify_commits_by_type_and_scope(v["commits"])
            v["commits"] = classified_commits
        self._display_number_of_commits_found_for_changelog(
            versioned_commits=versioned_commits
        )
        return versioned_commits

    def _separate_jama_ref(self, commit: Commit) -> Tuple[List[str], str]:
        """Separate jama reference from commit description

        :param commit: Commit object
        :type commit: Commit
        :return: Tuple of description without jama reference and jama reference
        :rtype: Tuple[List[str], str]
        """
        body_without_jama_ref = []
        jama_ref = ""
        if commit.body:
            for line in commit.body:
                if line.startswith(JAMA_TAG):
                    jama_ref = line.replace(JAMA_TAG, "").strip()
                else:
                    body_without_jama_ref.append(line)
        if jama_ref:
            return body_without_jama_ref, jama_ref
        return commit.body, ""

    def _append_title_and_description_with_matched_issue(
        self, commits: List[str], commit: Commit, issue_number: str
    ) -> None:
        """Append title without type and scope for feat and fixbug commit type,
        append title for other commit type. Append description if any
        and append issue number and url if issue number is found"""
        if commit.type == "feat" or commit.type == "fixbug":
            commit_title = commit.title_without_type_scope
        elif commit.type == "fixdefect":
            commit_title = self._format_fixdefect_title(commit=commit)
        else:
            commit_title = commit.title

        description, jama_ref = self._separate_jama_ref(commit=commit)

        if not issue_number:
            commits.append(
                {
                    "title": commit_title,
                    "description": description,
                    **({"jama_ref": jama_ref} if jama_ref else {}),
                }
            )
        else:
            url = f"{self.git.get_remote_origin_url()}/-/issues/{issue_number}"
            commits.append(
                {
                    "title": commit_title,
                    "description": description,
                    **({"jama_ref": jama_ref} if jama_ref else {}),
                    "gitlab_issue_id": issue_number,
                    "gitlab_issue_url": url,
                }
            )

    def _format_fixdefect_title(self, commit: Commit) -> str:
        """Format fixdefect title with the associated jama reference"""
        return (
            "**"
            + commit.scope.replace("(", "").replace(")", "")
            + "**"
            + ": "
            + commit.title_without_type_scope
        )

    def _classify_commits_by_type_and_scope(
        self, commits: List[Commit]
    ) -> Dict[str, Union[Dict[str, List[str]], List[str]]]:
        """Classify commits by type and scope and exclude merge commits"""
        (
            commits_fixdefect,
            commits_feat_add,
            commits_feat_change,
            commits_feat_remove,
            commits_fixbug,
            commits_other,
            commits_unspecified,
        ) = (
            [],
            [],
            [],
            [],
            [],
            [],
            [],
        )
        for commit in commits:
            match (commit.type, commit.scope):
                case ("fixdefect", commit.scope):
                    self._append_title_and_description_with_matched_issue(
                        commits=commits_fixdefect,
                        commit=commit,
                        issue_number=commit.mr_details.associated_closed_issue,
                    )
                case ("feat", "(add)"):
                    self._append_title_and_description_with_matched_issue(
                        commits=commits_feat_add,
                        commit=commit,
                        issue_number=commit.mr_details.associated_closed_issue,
                    )
                case ("feat", "(change)"):
                    self._append_title_and_description_with_matched_issue(
                        commits=commits_feat_change,
                        commit=commit,
                        issue_number=commit.mr_details.associated_closed_issue,
                    )
                case ("feat", "(remove)"):
                    self._append_title_and_description_with_matched_issue(
                        commits=commits_feat_remove,
                        commit=commit,
                        issue_number=commit.mr_details.associated_closed_issue,
                    )
                case ("fixbug", None):
                    self._append_title_and_description_with_matched_issue(
                        commits=commits_fixbug,
                        commit=commit,
                        issue_number=commit.mr_details.associated_closed_issue,
                    )
                case (_, _) if commit.type is not None:
                    self._append_title_and_description_with_matched_issue(
                        commits=commits_other,
                        commit=commit,
                        issue_number=commit.mr_details.associated_closed_issue,
                    )
                case (_, _) if commit.type is None:
                    if not commit.is_gitlab_merge_commit():
                        self._append_title_and_description_with_matched_issue(
                            commits=commits_unspecified,
                            commit=commit,
                            issue_number=commit.mr_details.associated_closed_issue,
                        )
        return self._create_commit_dict(
            commits_fixdefect=commits_fixdefect,
            commits_feat_add=commits_feat_add,
            commits_feat_change=commits_feat_change,
            commits_feat_remove=commits_feat_remove,
            commits_fixbug=commits_fixbug,
            commits_other=commits_other,
            commits_unspecified=commits_unspecified,
        )

    def _create_commit_dict(
        self,
        commits_fixdefect: List[str],
        commits_feat_add: List[str],
        commits_feat_change: List[str],
        commits_feat_remove: List[str],
        commits_fixbug: List[str],
        commits_other: List[str],
        commits_unspecified: List[str],
    ) -> Dict[str, Union[Dict[str, List[str]], List[str]]]:
        result = {}
        if commits_fixdefect:
            result["fixdefect"] = commits_fixdefect
        if commits_feat_add or commits_feat_change or commits_feat_remove:
            result["feature"] = {}
            if commits_feat_add:
                result["feature"]["added"] = commits_feat_add
            if commits_feat_change:
                result["feature"]["changed"] = commits_feat_change
            if commits_feat_remove:
                result["feature"]["removed"] = commits_feat_remove
        if commits_fixbug:
            result["fixbug"] = commits_fixbug
        if commits_other:
            result["other"] = commits_other
        if commits_unspecified:
            result["unspecified"] = commits_unspecified
        return result

    def _display_number_of_commits_found_for_changelog(
        self, versioned_commits: List[Dict[str, List[str]]]
    ) -> None:
        total = 0
        for v in versioned_commits:
            total += len(v.get("commits", {}).get("fixdefect", []))
            total += len(v.get("commits", {}).get("feature", {}).get("added", []))
            total += len(v.get("commits", {}).get("feature", {}).get("changed", []))
            total += len(v.get("commits", {}).get("feature", {}).get("removed", []))
            total += len(v.get("commits", {}).get("fixbug", []))
            total += len(v.get("commits", {}).get("other", []))
            total += len(v.get("commits", {}).get("unspecified", []))
        info(msg=f"Found {total} commits")



import os

import typer
from dotenv import load_dotenv

from lupin_grognard.__init__ import __version__
from lupin_grognard.core.check import (
    check_commit,
    check_max_allowed_major_commits,
    check_same_author_from_current_branch,
)
from lupin_grognard.core.commit.commit import add_additional_commit_info
from lupin_grognard.core.commit.commit_validator import (
    define_commits_check_mode,
    define_permissive_mode,
    CommitCheckModes,
)
from lupin_grognard.core.config import MAIN_BRANCHES_LIST
from lupin_grognard.core.doc_generator.changelog import Changelog
from lupin_grognard.core.doc_generator.reviewlog import Reviewlog
from lupin_grognard.core.doc_generator.ros2_docs import Ros2Docs
from lupin_grognard.core.git import Git
from lupin_grognard.core.gitlab import get_ci_mr_target_branch
from lupin_grognard.core.tools.ros2.package import find_ros_packages

from .core.tools.log_utils import die, info, warn
from .core.tools.utils import (
    configure_logging,
    display_current_branch_name,
    display_number_of_commits_to_check,
    display_supported_commit_types,
    generate_commit_list,
    generate_commits_range_to_check,
    get_current_branch_name,
)

load_dotenv()
GROG_BRANCHES = os.getenv("GROG_BRANCHES")
GROG_MAX_ALLOWED_COMMITS = os.getenv("GROG_MAX_ALLOWED_COMMITS")
GROG_DONT_CHECK_FOR_APPROVERS = os.getenv("GROG_DONT_CHECK_FOR_APPROVERS")
GROG_CLANG_FORMAT = os.getenv("GROG_CLANG_FORMAT")


cli = typer.Typer()


def tag_option() -> typer.Option:
    return typer.Option(
        None, "--from-tag", "-ft", help="Tag to start listing commits from (optional)"
    )


def output_option() -> typer.Option:
    return typer.Option(
        os.getcwd(),
        "--output",
        "-o",
        help="Directory to save the log file (default: current directory)",
    )


@cli.command()
def version():
    print(f"Version: {__version__}")


@cli.command()
def check_commits(
    CHECK_ALL_COMMITS: bool = typer.Option(
        False, "--all", "-a", help="check all commits from initial commit"
    ),
    grog_max_allowed_commits: int = typer.Option(
        1, "--grog-max-allowed-commits", "-max", envvar="GROG_MAX_ALLOWED_COMMITS"
    ),
    permissive_mode: bool = typer.Option(
        False, "--permissive", "-p", help="ignore command failure"
    ),
    no_approvers: bool = typer.Option(
        False, "--no-approvers", "-na", help="ignore approver check"
    ),
    tag: str = tag_option(),
):
    """
    Supported commit types: build, bump, ci, deps(add|change|remove), docs, enabler,
    feat(add|change|remove), fixbug, fixdefect(JAMA-xxx), refactor, test.
    Only one major commit types allowed per branch: "enabler", "feat", "fixbug", fixdefect or "refactor".

    Check every commit message since the last "merge request" in any of the branches in the
    main_branches_list : "main", "master", "dev", "develop", "development"

    - With --all option :
    grog check-commits [--all or -a] to check all commits from initial commit.
    This option is automatically set if current branch is a main one.

    - With --grog-max-allowed-commits option :
    grog check-commits [--grog-max-allowed-commits or -max] {int} to set the maximum number
    of commits allowed to the branch.
    Example : grog check-commits --grog-max-allowed-commits 10

    - With --permissive option :
    grog check-commits [--permissive or -p] to ignore command failure.
    This option is ignored for the 'main' and 'master' branches.

    - With --no-approvers option :
    grog check-commits [--no-approvers or -na] to ignore approver check.

    You can set GROG_MAX_ALLOWED_COMMITS env var in .env, .gitlab-ci.yml, gitlab...
    """
    configure_logging()
    git = Git()
    git.ensure_valid_git_repository()

    current_branch_name = get_current_branch_name()
    ci_mr_target_branch_name = get_ci_mr_target_branch()

    check_mode = define_commits_check_mode(
        current_branch=current_branch_name,
        ci_mr_target_branch=ci_mr_target_branch_name,
        CHECK_ALL_COMMITS_flag=CHECK_ALL_COMMITS,
    )

    permissive_mode = define_permissive_mode(
        check_mode=check_mode,
        permissive_mode=permissive_mode,
    )

    if CHECK_ALL_COMMITS:
        info(msg="Processing all commits since initial commit as '--all' option is set")
        git.ensure_not_shallow_clone()
    if no_approvers:
        warn(msg="Approver check disabled as '--no-approvers' option is set")

    if check_mode == CommitCheckModes.CHECK_ALL_COMMITS:
        grog_max_allowed_commits = 0
        git_log = git.get_log(tag=tag)
        commits = generate_commit_list(git_log.stdout)

        if ci_mr_target_branch_name in MAIN_BRANCHES_LIST:
            info(
                msg=(
                    "Processing check-commits for a pipeline merge request result "
                    f"to a '{ci_mr_target_branch_name}' target branch"
                )
            )
            commits = generate_commits_range_to_check(
                branch_list=MAIN_BRANCHES_LIST,
                commits=commits,
                ci_mr_target_branch_name=ci_mr_target_branch_name,
            )

    elif check_mode == CommitCheckModes.CHECK_CURRENT_BRANCH_ONLY:
        git_log = git.get_log(max_line_count=50, first_parent=True)
        commits = generate_commit_list(git_log.stdout)
        commits = generate_commits_range_to_check(
            branch_list=MAIN_BRANCHES_LIST,
            commits=commits,
        )

    if check_max_allowed_major_commits(
        commits=commits,
        major_commit_limit=grog_max_allowed_commits,
    ):
        display_supported_commit_types()
        display_current_branch_name(current_branch_name=current_branch_name)
        display_number_of_commits_to_check(commits=commits)

        if check_mode == CommitCheckModes.CHECK_CURRENT_BRANCH_ONLY:
            check_same_author_from_current_branch(commits=commits)

        check_commit(
            commits=commits,
            check_mode=check_mode,
            permissive_mode=permissive_mode,
            no_approvers=no_approvers,
        )


@cli.command()
def changelog(output: str = output_option(), tag: str = tag_option()):
    """Generate changelog
    Run this command `grog changelog` from a git repository to generate changelog.md file in the current directory.

    - With --output option :
    grog changelog [--output or -o] {string} to set the output to create changelog file.

    - With --from-tag option :
    grog changelog [--from-tag or -ft] {string} to set the tag to start listing commits from.
    """
    configure_logging()
    git = Git()
    git.ensure_valid_git_repository()
    git_log = git.get_log(tag=tag)
    if git_log.stderr:
        die(f"git error {git_log.return_code}, {git_log.stderr}")
    commits = generate_commit_list(commits_string=git_log.stdout)
    commits = add_additional_commit_info(commits=commits)
    Changelog(commits=commits).generate(output)


@cli.command()
def reviewlog(output: str = output_option(), tag: str = tag_option()):
    """Generate REVIEWLOG.html
    Run this command `grog reviewlog` from a git repository to generate REVIEWLOG.html file in the current directory.

    - With --output option :
    grog reviewlog [--output or -o] {string} to set the output to create reviewlog file.

    - With --from-tag option :
    grog changelog [--from-tag or -ft] {string} to set the tag to start listing commits from.
    """
    configure_logging()
    git = Git()
    git.ensure_valid_git_repository()
    git_log = git.get_log(tag=tag)
    if git_log.stderr:
        die(f"git error {git_log.return_code}, {git_log.stderr}")
    commits = generate_commit_list(commits_string=git_log.stdout)
    commits = add_additional_commit_info(commits=commits)
    Reviewlog(commits=commits).generate(output)


@cli.command()
def ros2docs(
    path: str = typer.Option(
        ..., "--path", "-p", help="path to search for ROS2 packages"
    )
):
    """Generate ROS2 documentation"""
    configure_logging()

    ros_packages = find_ros_packages(path)
    for path in ros_packages:
        api_doc = Ros2Docs(path=path)
        api_doc.generate_api_docs()



import unicodedata
from typing import Dict, List

from lupin_grognard.core.commit.commit import Commit, classify_commits_by_version
from lupin_grognard.core.doc_generator.jinja_generator import JinjaGenerator
from lupin_grognard.core.git import Git
from lupin_grognard.core.tools.log_utils import info


class Reviewlog(JinjaGenerator):
    def __init__(self, commits: List[Commit]):
        self.commits = commits
        self.git = Git()

    def generate(self, path: str):
        """Generate the reviewlog"""
        self.git.ensure_not_shallow_clone()

        project_url = self.git.get_remote_origin_url()
        project_name = project_url.split("/")[-1]
        info(msg=f"Collecting approvers report from {project_name}")
        versioned_approvers_report = self._get_approvers_report_by_version()
        approvers_participants = self._get_approvers_participants_in_versioned_report(
            versioned_approvers_report
        )
        self._generate_file(
            path=path,
            file_name="reviewlog.html",
            context={
                "approvers_report": versioned_approvers_report,
                "project_name": project_name,
                "project_url": project_url,
                "participants": approvers_participants,
            },
        )

    def _generate_file(self, path: str, file_name: str, context: Dict) -> None:
        return super()._generate_file(path, file_name, context)

    def _normalize_string(self, string: str) -> str:
        return (
            unicodedata.normalize("NFD", string)
            .encode("ascii", "ignore")
            .decode("utf-8")
        )

    def _remove_duplicate_participants(self, participants: List[str]) -> List[str]:
        """
        Remove duplicate participants from a list of participants
        :param participants: List of participants
        :return: List of participants with accent without duplicates

        Example:
        ["Cédric", "Cedric", "Aurelien", "Aurélien", "John Doe", "John Doe"] -> ["Aurélien", "Cédric", "John Doe"]
        """
        result = []
        participants_with_accents = set()
        participants_without_accents = set()
        for participant in participants:
            string_without_accents = self._normalize_string(participant)
            if string_without_accents == participant:
                participants_without_accents.add(participant)
            else:
                participants_with_accents.add(participant)

        for participant in participants_with_accents:
            normalized_string = self._normalize_string(participant)
            if normalized_string in participants_without_accents:
                participants_without_accents.remove(normalized_string)

        result = list(participants_without_accents)
        result.extend((list(participants_with_accents)))
        result = sorted(result)
        return [participant.title() for participant in result]

    def _get_approvers_participants_in_versioned_report(
        self, versioned_approvers_report: List[Dict[str, List[str]]]
    ) -> List[str]:
        approvers_participants = []
        for version in versioned_approvers_report:
            for report in version["commits"]:
                approvers = report.get("approvers", [])
                for approver in approvers:
                    if approver not in approvers_participants:
                        approvers_participants.append(approver)
        return self._remove_duplicate_participants(approvers_participants)

    def _get_name_without_mail_for_approvers(self, approvers: List[str]) -> List[str]:
        approvers_name = []
        for approver in approvers:
            approver_list = approver.split(" ")
            if "@" in approver_list[-1]:  # Mail always at the end in gitlab
                approver_name = " ".join(approver_list[:-1])
            else:
                approver_name = approver
            approvers_name.append(approver_name)
        return approvers_name

    def _get_approvers_report_by_version(self) -> List[dict]:
        classifed_commits = classify_commits_by_version(self.commits, self.git)
        for version in classifed_commits:
            approvers_report = []
            for commit in version["commits"]:
                if commit.mr_details.associated_closed_issue:
                    info(
                        msg=f"Collecting report for issue {commit.mr_details.associated_closed_issue}"
                    )
                    approvers = self._get_name_without_mail_for_approvers(
                        commit.mr_details.associated_approvers
                    )
                    approvers_report.append(
                        {
                            "commit_hash": commit.hash[:6],
                            "gitlab_issue_id": commit.mr_details.associated_closed_issue,
                            "title": commit.title,
                            "description": commit.body,
                            "autor": commit.author,
                            "date": commit.author_date,
                            "approvers": approvers,
                            "approver_date": commit.mr_details.associated_approvers_date,
                        }
                    )
            version["commits"] = approvers_report
        return classifed_commits



import os

from jinja2 import (
    Environment,
    PackageLoader,
    Template,
    TemplateError,
    TemplateNotFound,
    TemplateRuntimeError,
    select_autoescape,
)

from lupin_grognard.core.tools.log_utils import die, info
from lupin_grognard.core.tools.utils import write_file


class JinjaGenerator:
    def __init__(self, *args):
        super().__init__(*args)

    def _get_local_template(self, template_name: str) -> Template:
        try:
            package_loader = PackageLoader("lupin_grognard", "templates")
            env = Environment(
                loader=package_loader,
                autoescape=select_autoescape(),
                trim_blocks=True,  # Removes unnecessary spaces before and after blocks and loop
                lstrip_blocks=True,  # Removes unnecessary spaces before blocks and loop
            )
            return env.get_template(template_name)
        except TemplateNotFound:
            die(msg=f"Template 'lupin_grognard/templates/{template_name}' not found")

    def _replace_extension_to_j2(self, file_name: str) -> str:
        assert file_name.count(".") == 1
        base_name = file_name.rsplit(".", 1)[0]
        return f"{base_name}.j2"

    def _generate_file(self, path: str, file_name: str, context={}) -> None:
        template_name = self._replace_extension_to_j2(file_name=file_name)
        template = self._get_local_template(template_name=template_name)
        try:
            content = template.render(context)
        except (TemplateError, TemplateRuntimeError) as e:
            die(msg=f"Error rendering Jinja2 template: {e}")
        if not path:
            info(msg=f"Generating {file_name} file")
            write_file(file=f"{file_name}", content=content)
        else:
            self._generate_path(path=path)
            info(msg=f"Generating {path}/{file_name} file")
            write_file(file=f"{path}/{file_name}", content=content)
        info(msg="File generated")

    def _generate_path(self, path: str) -> None:
        if not os.path.exists(path):
            path = path[:-1] if path.endswith(os.path.sep) else path
            os.makedirs(path)



MAIN_BRANCHES_LIST = ["main", "master", "dev", "develop", "development"]

INITIAL_COMMIT = [
    "Initial commit",  # Generated by GitLab
    "Add README.md",  # Generated by GitKraken
]

COMMIT_DELIMITER = "--commit-delimiter--"

MIN_COMMIT_DESCR_LENTH = 10

MAX_COMMIT_DESCR_LINES = 20

SUCCESS = "Commit validation: success!"

FAILED = "Commit validation: failed!"

TITLE_FAILED = "Please enter a commit message in the conventional commits format."

BODY_FAILED = (
    "The commit description contains a conventional commit prefix "
    f"or a line of less than {MIN_COMMIT_DESCR_LENTH} characters.\n"
    "Did you forget to edit and clean up the commit message after a squash?"
)

MERGE_FAILED = "Commit merges should have at least one approver."

EMOJI_CHECK = ":check_mark_button:"
EMOJI_CROSS = ":cross_mark:"

COMMIT_TYPE_MUST_HAVE_SCOPE = {
    "fixdefect": (
        "The defect being fixed must be given in argument with the syntax 'fixdefect(JAMA-xxx)' "
        "where 'xxx' is the ID of the associated item in JAMA"
    ),
    "other": (
        "Commits of type '{c_type}' must be of the form '{c_type}(add)', "
        "'{c_type}(change)' or '{c_type}(remove)'"
    ),
}

COMMIT_TYPE_MUST_NOT_HAVE_SCOPE = (
    "Commits of type '{}' must not specify a scope but simply use the form '{}'"
)

PATTERN = (
    r"(?s)"  # To explictly make . match new line
    r"(build|bump|ci|deps|docs|enabler|feat|fixbug|fixdefect|refactor|test)"  # type
    r"(\(\S+\))?!?:"  # scope
    r"( [^\n\r]+)"  # subject
    r"((\n\n.*)|(\s*))?$"
)

MAJOR_COMMIT_TYPES = ["enabler", "feat", "fixbug", "fixdefect", "refactor"]

COMMIT_WITH_SCOPE = ["deps", "feat", "fixdefect"]

JAMA_REGEX = r"SmlPrep-SUBSR-\d+"

JAMA_FIXDEFECT_REGEX = r"\(JAMA-[a-zA-Z0-9-]+\)"

JAMA_TAG = "JAMA:"

IMPACT_TAG = "IMPACT:"



import re
import string
from enum import Enum
from typing import List, Tuple

from lupin_grognard.core.commit.commit import Commit
from lupin_grognard.core.commit.commit_error import BodyError, ErrorCount
from lupin_grognard.core.commit.commit_reporter import CommitReporter
from lupin_grognard.core.config import (
    COMMIT_TYPE_MUST_HAVE_SCOPE,
    COMMIT_TYPE_MUST_NOT_HAVE_SCOPE,
    COMMIT_WITH_SCOPE,
    INITIAL_COMMIT,
    IMPACT_TAG,
    JAMA_FIXDEFECT_REGEX,
    JAMA_REGEX,
    JAMA_TAG,
    MAIN_BRANCHES_LIST,
    MAX_COMMIT_DESCR_LINES,
    MIN_COMMIT_DESCR_LENTH,
    PATTERN,
    TITLE_FAILED,
)


class CommitCheckModes(Enum):
    # Check all commits starting from the initial commit
    CHECK_ALL_COMMITS = 1
    # Check all commits of the current branch (until the last merge)
    CHECK_CURRENT_BRANCH_ONLY = 2


def define_commits_check_mode(
    current_branch: str, ci_mr_target_branch: str, CHECK_ALL_COMMITS_flag: bool
) -> CommitCheckModes:
    if (
        current_branch in MAIN_BRANCHES_LIST
        or ci_mr_target_branch in MAIN_BRANCHES_LIST
        or CHECK_ALL_COMMITS_flag
    ):
        return CommitCheckModes.CHECK_ALL_COMMITS
    else:
        return CommitCheckModes.CHECK_CURRENT_BRANCH_ONLY


def define_permissive_mode(
    check_mode: CommitCheckModes,
    permissive_mode: bool,
) -> bool:
    """
    Ensures that permissive mode can be enabled on a CommitCheckModes.CHECK_ALL_COMMITS
    if the --permissive flag is specified.

    Args:
        check_mode (CommitCheckModes): The check mode.
        permissive_mode (bool): The permissive flag.
    """
    return check_mode == CommitCheckModes.CHECK_ALL_COMMITS and permissive_mode


class CommitValidator(Commit):
    def __init__(
        self,
        commit: Commit,
        error_counter: ErrorCount,
        check_mode: CommitCheckModes,
        no_approvers: bool = False,
    ):
        super().__init__(commit=commit.commit)
        self.reporter = CommitReporter(commit=commit)
        self.error_counter = error_counter
        self.check_mode = check_mode
        self.no_approvers = no_approvers

    def perform_checks(self) -> None:
        if self.check_mode == CommitCheckModes.CHECK_ALL_COMMITS:
            if self.is_gitlab_merge_commit():
                if not self.no_approvers:
                    if not self._validate_commit_merge():
                        self.error_counter.increment_merge_error()
                else:
                    self.reporter.display_merge_report_with_no_approver_option()
            else:
                if not self._validate_commit_title():
                    self.error_counter.increment_title_error()
                if not self._validate_body():
                    self.error_counter.increment_body_error()

        if self.check_mode == CommitCheckModes.CHECK_CURRENT_BRANCH_ONLY:
            if not self._validate_commit_title():
                self.error_counter.increment_title_error()
            if not self.is_gitlab_merge_commit():
                if not self._validate_body():
                    self.error_counter.increment_body_error()

    def _validate_commit_title(self) -> bool:
        if self._validate_commit_message(self.title, self.type, self.scope):
            self.reporter.display_valid_title_report()
            return True
        else:
            return False

    def _validate_body(self) -> bool:
        self.body_error = BodyError(
            is_conventional=[],
            descr_is_too_short=[],
            num_empty_line=0,
            invalid_body_length=False,
            duplicate_jama_refs=[],
            invalid_jama_refs=False,
            impact_tag_is_missing=False,
            impact_content_line_is_short=False,
        )

        self._validate_tags_in_description()

        if self.body:
            if not self._is_commit_body_length_valid():
                self.body_error.invalid_body_length = True

            for message in self.body:
                if self._is_conventional_commit_body_valid(message=message):
                    self.body_error.is_conventional.append(
                        message
                    )  # must not start with a conventional message
                if not self._is_commit_body_line_length_valid(message=message):
                    if message != "":
                        self.body_error.descr_is_too_short.append(message)
                    else:
                        self.body_error.num_empty_line += 1
        if any(
            [
                self.body_error.is_conventional,
                self.body_error.descr_is_too_short,
                self.body_error.num_empty_line > 0,
                self.body_error.invalid_body_length,
                self.body_error.duplicate_jama_refs,
                self.body_error.invalid_jama_refs,
                self.body_error.impact_tag_is_missing,
                self.body_error.impact_content_line_is_short,
            ]
        ):
            self.reporter.display_body_report(self.body_error)
            return False
        return True

    def _validate_tags_in_description(self):
        if self._check_if_tag_exists(JAMA_TAG):
            self._validate_jama_referencing()

        if self.type == "fixbug":
            self._validate_impact_referencing_for_fixbug()

    def _check_if_tag_exists(self, tag: str) -> bool:
        if self.body:
            return any([line.startswith(tag) for line in self.body])

    def _validate_jama_referencing(self) -> None:
        """
        Validate JAMA referencing in the body text

        - Check if the signature line starts with a valid JAMA reference
        - Validate individual JAMA items by detecting duplicates and invalid references
        """
        (
            duplicate_jama_refs,
            invalid_jama_refs,
        ) = self._validate_jama_items()
        if duplicate_jama_refs:
            self.body_error.duplicate_jama_refs = duplicate_jama_refs
        if invalid_jama_refs:
            self.body_error.invalid_jama_refs = invalid_jama_refs

    def _validate_impact_referencing_for_fixbug(self) -> None:
        """
        Validate 'impact' referencing in the body text for fixbug commits type.

        - Check if the signature line starts with 'IMPACT:'
        - Check if the signature line is more than 10 characters long
        """
        if self._check_if_tag_exists(IMPACT_TAG):
            impact_ref_line = self._get_line_starting_with_tag(IMPACT_TAG)
            impact_line_content = impact_ref_line.replace(IMPACT_TAG, "").strip()
            if len(impact_line_content) <= 10:
                self.body_error.impact_content_line_is_short = True
        else:
            self.body_error.impact_tag_is_missing = True

    def _get_line_starting_with_tag(self, tag: str) -> str:
        """
        Get line starting with a given tag

        Args:
            tag (str): The tag to check for

        Returns:
            str: The line starting with the given tag
        """
        if self.body:
            for line in self.body:
                if line.startswith(tag):
                    return line
        return ""

    def _validate_jama_items(self) -> Tuple[List[str], List[str]]:
        """
        Validate individual JAMA items
        - Check if the JAMA item matches the JAMA reference pattern defined by JAMA_REGEX
        - Check for duplicate JAMA items and record them
        - Check for invalid JAMA items and record them

        Args:
            jama_ref_line (str): The line starting with 'JAMA:'

        Returns:
            A tuple containing two lists:
            - duplicate_jama_refs (List[str]): List of duplicate JAMA items
            - invalid_jama_refs (List[str]): List of invalid JAMA items
        """
        jama_ref_line = self._get_line_starting_with_tag(JAMA_TAG)
        content_jama_ref_line = jama_ref_line.replace(JAMA_TAG, "").strip()
        jama_refs = [
            ref for ref in re.split(r"[\s,]+", content_jama_ref_line) if ref != ""
        ]
        unique_jama_refs = set()
        duplicate_jama_refs = []
        invalid_jama_refs = []

        for jama_ref in jama_refs:
            if not re.match(JAMA_REGEX, jama_ref):
                invalid_jama_refs.append(jama_ref)
            elif jama_ref in unique_jama_refs and jama_ref not in duplicate_jama_refs:
                duplicate_jama_refs.append(jama_ref)
            else:
                unique_jama_refs.add(jama_ref)
        return duplicate_jama_refs, invalid_jama_refs

    def _validate_commit_message(
        self, commit_msg: str, c_type: str, scope: str
    ) -> bool:
        if self._is_special_commit(commit_msg=commit_msg):
            return True

        match c_type:
            case None:
                self.reporter.display_invalid_title_report(error_message=TITLE_FAILED)
                return False
            case match_type if (match_type := c_type) in COMMIT_WITH_SCOPE:
                return self._validate_commit_message_for_specific_type(
                    scope=scope, c_type=match_type
                )
            case _:
                return self._validate_commit_message_for_generic_type(
                    c_type=c_type, scope=scope
                )

    def _is_conventional_commit_body_valid(self, message: str) -> bool:
        """Checks if the line in the body of a commit message starts with a conventional commit"""
        return bool(re.match(PATTERN, message))

    def _is_commit_body_line_length_valid(self, message: str) -> bool:
        """Checks if the body line is not less than MIN_COMMIT_DESCR_LENTH"""
        return len(message) >= MIN_COMMIT_DESCR_LENTH

    def _is_commit_body_length_valid(self) -> bool:
        """Checks if the body length is not greater than MAX_COMMIT_DESCR_LINES"""
        return len(self.body) <= MAX_COMMIT_DESCR_LINES

    def _is_special_commit(self, commit_msg: str) -> bool:
        """Checks if the commit is a Merge or in the list of initial commits"""
        return commit_msg.startswith("Merge") or commit_msg in INITIAL_COMMIT

    def _validate_commit_merge(self) -> bool:
        if len(self.approvers) < 1:
            self.reporter.display_merge_report_no_approver()
            return False
        else:
            if self._is_mr_approver_also_author():
                return False
            else:
                self.reporter.display_valid_merge_report(approvers=self.approvers)
                return True

    def _is_mr_approver_also_author(self) -> bool:
        """
        Check if the MR approver is also an author of the commit(s) associated with the MR.
        """
        rejected_user_emails = self._find_mr_approvers_that_are_also_authors()
        if rejected_user_emails:
            self.reporter.display_merge_approver_same_author_child_commit(
                rejected_user_emails
            )
            return True
        return False

    def _find_mr_approvers_that_are_also_authors(self) -> list[str]:
        """Find MR approvers that are also authors of the commit(s) associated with the MR."""
        rejected_user_emails = []
        for author in self.mr_details.associated_commits_authors:
            if author in self.approvers_mail:
                rejected_user_emails.append(author)
        return rejected_user_emails

    def _validate_commit_message_for_specific_type(
        self, scope: str, c_type: str
    ) -> bool:
        """
        Validates the scope for a COMMIT_WITH_SCOPE list.

        Args:
            scope (str): The scope of the commit message
            c_type (str): The commit type

        Returns:
            bool: True if the commit message is valid, False otherwise
        """
        if c_type == "fixdefect":
            if scope is None or not re.match(JAMA_FIXDEFECT_REGEX, scope):
                self.reporter.display_invalid_title_report(
                    error_message=COMMIT_TYPE_MUST_HAVE_SCOPE[c_type]
                )
                return False
        elif scope is None or scope not in ["(add)", "(change)", "(remove)"]:
            self.reporter.display_invalid_title_report(
                error_message=COMMIT_TYPE_MUST_HAVE_SCOPE["other"].format(c_type=c_type)
            )
            return False
        return True

    def _remove_punctuation_around_words(self, words: List[str]) -> List[str]:
        """Removes punctuation around words"""
        return [word.strip(string.punctuation) for word in words if word != ""]

    def _validate_commit_message_for_generic_type(self, c_type, scope: str) -> bool:
        """Validates other commit types do not contain a scope"""
        if scope is None:
            return True
        else:
            error_message = COMMIT_TYPE_MUST_NOT_HAVE_SCOPE.format(c_type, c_type)
            self.reporter.display_invalid_title_report(error_message=error_message)
            return False



import re
from datetime import datetime
from typing import Dict, List, Optional, Tuple, Union

from lupin_grognard.core.config import (
    INITIAL_COMMIT,
    MAJOR_COMMIT_TYPES,
    PATTERN,
)
from lupin_grognard.core.git import Git
from lupin_grognard.core.tools.log_utils import info, warn


class GitLabMergeRequestDetails:
    def __init__(self):
        self.associated_closed_issue: Optional[str] = None
        self.associated_approvers: Optional[str] = None
        self.associated_approvers_date: Optional[str] = None
        self.associated_commits_authors: Optional[List[str]] = list()


class Commit:
    def __init__(self, commit: str):
        self.commit = commit
        self.parents: Optional[List[str]] = None
        self.mr_details: Optional[GitLabMergeRequestDetails] = (
            GitLabMergeRequestDetails()
        )

    @property
    def hash(self) -> str:
        return self._extract(start="hash>>")

    @property
    def author(self) -> str:
        return self._extract(start="author>>")

    @property
    def author_mail(self) -> str:
        return self._extract(start="author_mail>>")

    @property
    def author_date(self) -> str:
        timestamp = self._extract(start="author_date>>")
        date_object = datetime.fromtimestamp(int(timestamp))
        return date_object.strftime("%d/%m/%y %I:%M %p")

    @property
    def author_and_committer_mail(self):
        return Git().get_author_and_commiter_mail(self.hash)

    @property
    def title(self) -> str:
        return self._extract(start="title>>")

    @property
    def title_without_type_scope(self) -> str:
        """Returns commit title without type and scope"""
        start = self.title.find(":") + 1
        title = self.title[start:].strip()
        return title[0].upper() + title[1:]

    @property
    def type(self) -> str | None:
        """Returns the conventional commit type if present"""
        match = re.match(PATTERN, self.title)
        return match.groups()[0] if match else None

    @property
    def scope(self) -> str | None:
        """Returns the conventional commit scope if present"""
        match = re.match(PATTERN, self.title)
        return match.groups()[1] if match else None

    @property
    def body(self) -> List[str] | None:
        body = self._extract(start="body>>", end="<<body")
        if body == "":
            return None

        # remove last \n if present cause 'git commit -m "description"' or other git software adds it automatically
        body = body.rstrip("\n")

        return [
            self._remove_markdown_list_markers(message) for message in body.split("\n")
        ]

    @property
    def closes_issues(self) -> List | str:
        """Returns the list of issues closed by the commit"""
        if self.body:
            for line in self.body:
                if line.startswith("Closes #"):  # Closes #465, #190 and #400
                    return re.findall(r"#(\d+)", line)  # ['465', '190', '400']
        warn(f"Could not find the issue closed by the commit '{self.title}'")
        return "0"

    @property
    def approvers(self) -> List[str]:
        approvers = []
        if self.body:
            for line in self.body:
                if line.startswith("Approved-by: "):
                    approver = line.split("Approved-by: ")[1]
                    approver = approver.translate(str.maketrans("", "", "<>"))
                    approvers.append(approver)
            return approvers
        return list()

    @property
    def approvers_name(self) -> List[str]:
        if self.approvers:
            return [" ".join(approver.split(" ")[:-1]) for approver in self.approvers]
        return list()

    @property
    def approvers_mail(self) -> List[str]:
        if self.approvers:
            return [approver.split(" ")[-1] for approver in self.approvers]
        return list()

    def _extract(self, start: str, end: str = "\n") -> str:
        start_index = self.commit.find(start) + len(start)
        return self.commit[start_index : self.commit.find(end, start_index)]

    def is_gitlab_merge_commit(self) -> bool:
        return self.title.startswith("Merge branch")

    def is_major_commit(self) -> bool:
        """Returns if the commit is a major commit type"""
        return self.type in MAJOR_COMMIT_TYPES

    def is_not_initial_commit(self) -> bool:
        """Returns if the commit is not an initial commit"""
        return self.title not in INITIAL_COMMIT

    def _remove_markdown_list_markers(self, message: str) -> str:
        return message.lstrip("-* ")


def add_additional_commit_info(commits: List["Commit"]) -> List["Commit"]:
    """
    Returns a list of Commit objects with additional information such as closed issues, approvers,
    and date it was approved for each commit from associated merge request

    :param commits: List of Commit objects
    :return: List of Commit objects with additional information

    additional information:
        self.mr_details.associated_closed_issue = None if merge commit else "1"
        self.mr_details.associated_approvers = None if merge commit else "John Doe"
        self.mr_details.associated_approvers_date = None if merge commit else "10/03/23 06:48 PM"
        self.parrents = ["hash1", "hash2"] if merge commit else ["hash1"]
    """
    commits = get_parents_for_commits(commits=commits)
    merge_commits_hash, merge_commits_mapping = get_data_from_merge_commit(
        commits=commits
    )
    commits = add_associated_data_to_commit_from_merge(
        merge_commits_hash=merge_commits_hash,
        merge_commits_mapping=merge_commits_mapping,
        commits=commits,
    )
    return commits


def get_data_from_merge_commit(
    commits: List["Commit"],
) -> Tuple[List[str], Dict[str, Dict[str, Union[str, List[str], str]]]]:
    """
    Return a tuple containing a dictionary with information about merge commits and a list of merge commit hashes.

    The merge commit information dictionary has commit parent hash as keys, and the following information as values:
        - "mr_details.associated_closed_issue": L'id of the gitlab closed issue.
        - "approvers": A list of the usernames who approved the merge commit.
        - "date": The date the merge commit was approved.

    :param commits: A list of Commit objects.
    :type commits: List["Commit"]
    :return:
        - merge_commits_hash, merge_commits_mapping
        - A tuple containing the list of merge commit hashes and a merge commit information dictionary.
    :rtype: Tuple[List[str], Dict[str, Dict[str, Union[str, List[str], str]]]]
    """
    merge_commits_hash = []
    merge_commits_mapping = {}
    for commit in commits:
        if len(commit.parents) == 2:  # check if it is a merge commit
            merge_commits_hash.append(commit.hash)
            if commit.closes_issues:
                merge_commits_mapping[commit.parents[1]] = {
                    "close_issue": commit.closes_issues[0],
                    "approvers": commit.approvers,
                    "date": commit.author_date,
                }
    return merge_commits_hash, merge_commits_mapping


def add_associated_data_to_commit_from_merge(
    merge_commits_hash: List[str],
    merge_commits_mapping: Dict[str, Dict[str, Union[str, List[str], str]]],
    commits: List["Commit"],
) -> List["Commit"]:
    for commit in commits:
        if (
            len(commit.parents) == 1  # check if it is not a merge commit
            and commit.hash in merge_commits_mapping
            and commit.title not in INITIAL_COMMIT
        ):
            commit.mr_details.associated_closed_issue = merge_commits_mapping[
                commit.hash
            ]["close_issue"]
            commit.mr_details.associated_approvers = merge_commits_mapping[commit.hash][
                "approvers"
            ]
            commit.mr_details.associated_approvers_date = merge_commits_mapping[
                commit.hash
            ]["date"]
            if (
                commit.parents[0] not in merge_commits_mapping
                and commit.parents[0] not in merge_commits_hash
            ):  # the commit parent shares the same merge commit
                merge_commits_mapping[commit.parents[0]] = merge_commits_mapping[
                    commit.hash
                ]
    return commits


def fill_authors_for_mr_commits(commits: List["Commit"]) -> List["Commit"]:
    """
    For each merge commit in the list, fill the authors of its children commits
    (that are associated to the merge request).

    Args:
        commits (List[Commit]): The list of commits to check.

    Returns:
        List[Commit]: The modified list of commits with the authors added to each merge commit.
    """

    commits = get_parents_for_commits(commits=commits)

    for commit in commits:
        if commit.is_gitlab_merge_commit():
            add_author_to_merge_request(
                merge_request=commit, hash=commit.parents[1], commits=commits
            )
    return commits


def find_commit_by_hash(hash: str, commits: List["Commit"]) -> Commit:
    """Find a commit in a list of commits by its hash.

    Args:
        hash (str): The hash of the commit to retrieve.
        commits (List[Commit]): The list of commits to search in.

    Returns:
        Commit: The commit with the given hash.
    """
    for commit in commits:
        if commit.hash == hash:
            return commit


def add_author_to_merge_request(
    merge_request: Commit, hash: str, commits: List["Commit"]
) -> None:
    """
    Recursively adds the author of a child commit to the associated merge request.

    Args:
        merge_request (Commit): The merge request to add the author.
        hash (str): The hash of the child commit.
        commits (List[Commit]): The list of commits.

    Returns:
        None
    """
    child_commit = find_commit_by_hash(hash=hash, commits=commits)

    if not child_commit:
        return
    if (
        not child_commit.is_gitlab_merge_commit()
        and child_commit.is_not_initial_commit()
    ):
        merge_request.mr_details.associated_commits_authors.append(
            child_commit.author_mail
        )

        if len(child_commit.parents) == 1:
            add_author_to_merge_request(
                merge_request=merge_request,
                hash=child_commit.parents[0],
                commits=commits,
            )


def get_parents_for_commits(commits: List["Commit"]) -> List["Commit"]:
    """Get the parents for each commit in the list.

    Args:
        commits (List[Commit]): The list of commits.

    Returns:
        List[Commit]: The list of commits with the parents added.
    """
    for commit in commits:
        commit.parents = Git().get_parents(commit_hash=commit.hash)
    return commits


def classify_commits_by_version(
    commits: List[Commit], git: Git
) -> List[Dict[str, List[Commit]]]:
    """
    Classify a list of commits by version.

    Args:
        commits (List[Commit]): A list of Commit objects to classify.
        git (Git): An instance of the Git.

    Returns:
        List[Dict[str, List[Commit]]]: A list of dictionaries, each representing a version
        containing the commits associated with that version.

    This function iterates through a list of commits and classifies them based on the tags
    present in the provided Git instance. It groups the commits by version and returns
    a list of dictionaries where each dictionary contains information about a version
    along with the commits associated with that version.
    """
    versions = []
    current_version = {}
    tag_list = git.get_tags()
    for commit in commits:
        for tag in tag_list:
            if commit.hash in tag[1]:
                info(msg=f"Found tag {tag[0]}")
                commit_tag = tag[0]
                date_tag = tag[2]
                if current_version:
                    versions.append(current_version)
                current_version = {
                    "version": commit_tag,
                    "date": date_tag,
                    "commits": [],
                }
                current_version["commits"].append(commit)
                break
        else:
            if not current_version:
                current_version = {
                    "version": "Unreleased",
                    "date": "",
                    "commits": [],
                }
            current_version["commits"].append(commit)
    if current_version:
        versions.append(current_version)
    return versions



import subprocess
from typing import NamedTuple


class Command(NamedTuple):
    stdout: bytes
    stderr: bytes
    return_code: int


def remove_last_line_break(data: str) -> str:
    last_line_break = data.rfind("\n")
    if last_line_break == -1:
        return data
    else:
        return data[:last_line_break]


def _decode(data: bytes) -> str:
    data = data.decode("utf-8")
    return remove_last_line_break(data=data)


def run_command(command: str) -> Command:
    process = subprocess.Popen(
        command,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        stdin=subprocess.PIPE,
        shell=True,
    )
    stdout, stderr = process.communicate()
    return_code = process.returncode
    stdout = _decode(data=stdout)
    stderr = _decode(data=stderr)
    return Command(
        stdout,
        stderr,
        return_code,
    )



import os
from typing import Dict

from lupin_grognard.core.doc_generator.jinja_generator import JinjaGenerator
from lupin_grognard.core.tools.ros2.package import PackageRos2


class Ros2Docs(JinjaGenerator):
    def __init__(self, path: str):
        self.path = path
        self.package = PackageRos2(package_path=path)

    def get_package_info(self) -> Dict:
        """
        Returns a dictionary with the following keys:
        - package: PackageRos2 object
        - msgs: a list of Msg objects
        - srvs: a list of Srv objects
        - actions: a list of Action objects

        :return: a dictionary with the package info
        :rtype: Dict"""
        package = self.package
        return {
            "package": package,
            "msgs": package._find_files(file_type="msg"),
            "srvs": package._find_files(file_type="srv"),
            "actions": package._find_files(file_type="action"),
        }

    def generate_api_docs(self) -> None:
        if not self._is_docs_dir_exists():
            os.mkdir(os.path.join(self.path, "docs"))

        self._generate_file(
            path=os.path.join(self.path, "docs"),
            file_name="RosApi.md",
            context=self.get_package_info(),
        )

    def _is_docs_dir_exists(self) -> bool:
        return os.path.isdir(os.path.join(self.path, "docs"))

    def _generate_file(self, path: str, file_name: str, context: Dict) -> None:
        return super()._generate_file(path, file_name, context)



import os

from .parser import (
    ActionSpecification,
    MessageSpecification,
    ServiceSpecification,
    extract_file_level_comments,
    parse_action_file,
    parse_message_file,
    parse_service_file,
)


class InterfaceRos2:
    def __init__(self, package_name: str, path: str):
        self.package_name = package_name
        self.path = path

    @property
    def name(self) -> str:
        suffixes = (".msg", ".srv", ".action")
        for suffix in suffixes:
            if self.path.endswith(suffix):
                return os.path.basename(self.path)[: -len(suffix)]


class Msg(InterfaceRos2):
    def __init__(self, package_name: str, path: str):
        super().__init__(package_name, path)

    def parse_message_file(self) -> MessageSpecification:
        return parse_message_file(self.package_name, self.path)

    @property
    def comment(self) -> str:
        """
        The comment block at the top of the message file.
        """
        parsed_message = self.parse_message_file()
        comments_list = parsed_message.annotations.get("comment", [])
        return "\n".join(comments_list)

    @property
    def code(self) -> str:
        """
        The code block after the comment block in the message file.
        """
        with open(self.path, "r", encoding="utf-8") as f:
            message_string = f.read()
        code_list = extract_file_level_comments(message_string=message_string)[1]
        return "\n".join(code_list).strip()


class Srv(InterfaceRos2):
    def __init__(self, package_name: str, path: str):
        super().__init__(package_name, path)

    def parse_service_file(self) -> ServiceSpecification:
        return parse_service_file(self.package_name, self.path)

    @property
    def comment(self) -> str:
        """
        The comment block at the top of the service file.
        """
        parsed_service = self.parse_service_file()
        comments_list = parsed_service.request.annotations.get("comment", [])
        return "\n".join(comments_list)


class Action(InterfaceRos2):
    def __init__(self, package_name: str, path: str):
        super().__init__(package_name, path)

    def parse_action_file(self) -> ActionSpecification:
        return parse_action_file(self.package_name, self.path)

    @property
    def comment(self) -> str:
        """
        The comment block at the top of the action file.
        """
        parsed_action = self.parse_action_file()
        comments_list = parsed_action.goal.annotations.get("comment", [])
        return "\n".join(comments_list)



from typing import List

from emoji import emojize

from lupin_grognard.core.commit.commit import Commit
from lupin_grognard.core.commit.commit_error import BodyError
from lupin_grognard.core.config import EMOJI_CHECK, EMOJI_CROSS


class CommitReporter(Commit):
    def __init__(self, commit: Commit):
        super().__init__(commit=commit.commit)

    def display_valid_title_report(self) -> None:
        print(emojize(f"{EMOJI_CHECK} Commit {self.hash[:6]}: {self.title}"))

    def display_invalid_title_report(self, error_message) -> None:
        print(emojize(f"{EMOJI_CROSS} Commit {self.hash[:6]}: {self.title}"))
        print(f"   {error_message}")

    def display_body_report(self, body_error: BodyError) -> None:
        print(emojize(f"{EMOJI_CROSS} Error in message discription:"))

        for message in body_error.is_conventional:
            print(f"    The line can't start with a conventional commit: '{message}'")
        for message in body_error.descr_is_too_short:
            print(f"    Text is too short in the commit description: '{message}'")

        if body_error.num_empty_line == 1:
            print("    Found an empty line in the commit description")
        elif body_error.num_empty_line > 1:
            print(
                f"    Found {body_error.num_empty_line} empty lines in the commit description"
            )

        if body_error.invalid_body_length:
            print("    The description should not exceed 20 lines")

        if body_error.invalid_jama_refs:
            if len(body_error.invalid_jama_refs) == 1:
                print(
                    (
                        f"    Found that one JAMA reference was invalid: {body_error.invalid_jama_refs[0]}"
                    )
                )
            else:
                print(
                    (
                        "    Found that several JAMA references were "
                        f"invalid: {', '.join(body_error.invalid_jama_refs)}"
                    )
                )

        if body_error.duplicate_jama_refs:
            if len(body_error.duplicate_jama_refs) == 1:
                print(
                    (
                        f"    Found duplicate JAMA reference: {body_error.duplicate_jama_refs[0]}"
                    )
                )
            else:
                print(
                    (
                        f"    Found duplicate JAMA references: {', '.join(body_error.duplicate_jama_refs)}"
                    )
                )

        if body_error.impact_tag_is_missing:
            print(
                (
                    "    Commits of type 'fixbug' must contain in their description a line starting with 'IMPACT:' "
                    "that describes the impact of the fix."
                )
            )

        if body_error.impact_content_line_is_short:
            print("    The details provided in the 'IMPACT:' section is too short.")

    def display_merge_info(self) -> None:
        print(f"   - Merged on {self.author_date} by {self.author}, {self.author_mail}")
        print(
            f"   - Closes issue: {self.closes_issues if self.closes_issues else 'Not found'}"
        )
        print(f"   - Commit title: {self.title}")

    def display_valid_merge_report(self, approvers: List[str]) -> None:
        if len(approvers) == 1:
            print(
                emojize(
                    f"{EMOJI_CHECK} Merge commit {self.hash[:6]}: Approver: {approvers[0]}"
                )
            )
        else:
            many_pprovers = ", ".join(approvers[:-1]) + " and " + approvers[-1]
            print(
                emojize(
                    f"{EMOJI_CHECK} Merge commit {self.hash[:6]}: Approvers: {many_pprovers}"
                )
            )
        self.display_merge_info()

    def display_merge_report_with_no_approver_option(self) -> None:
        print(emojize(f"{EMOJI_CHECK} Merge commit {self.hash[:6]}:"))
        self.display_merge_info()

    def display_merge_approver_same_author_child_commit(
        self, author_child_commit_same_mr_approver
    ) -> None:
        if len(author_child_commit_same_mr_approver) > 1:
            print(
                emojize(
                    (
                        f"{EMOJI_CROSS} Merge commit {self.hash[:6]}: "
                        f"The authors of the child commit '{', '.join(author_child_commit_same_mr_approver)}' "
                        "also approves the merge commit."
                    )
                )
            )
        else:
            print(
                emojize(
                    (
                        f"{EMOJI_CROSS} Merge commit {self.hash[:6]}: "
                        f"The author of the child commits '{author_child_commit_same_mr_approver[0]}' "
                        "also approve the merge commit."
                    )
                )
            )
        self.display_merge_info()

    def display_merge_report_no_approver(self) -> None:
        print(emojize(f"{EMOJI_CROSS} Merge commit {self.hash[:6]}: No approver found"))
        self.display_merge_info()



import sys
from dataclasses import dataclass
from typing import List

from emoji import emojize

from lupin_grognard.core.config import (
    BODY_FAILED,
    FAILED,
    MERGE_FAILED,
    SUCCESS,
    TITLE_FAILED,
)


class ErrorCount:
    def __init__(self):
        self.body_error = 0
        self.title_error = 0
        self.merge_error = 0

    def increment_body_error(self):
        self.body_error += 1

    def increment_title_error(self):
        self.title_error += 1

    def increment_merge_error(self):
        self.merge_error += 1

    def error_report(self, permissive_mode: bool):
        if not (self.title_error + self.body_error + self.merge_error):
            print(SUCCESS)
        else:
            print(FAILED)
            print(
                f"Errors found: {self.title_error + self.body_error + self.merge_error}"
            )
            if self.title_error > 0:
                print(TITLE_FAILED)
            if self.body_error > 0:
                print(BODY_FAILED)
            if self.merge_error > 0:
                print(MERGE_FAILED)

            if permissive_mode:
                print(
                    emojize(
                        ":warning:  WARNING: ignoring command failure because running in permissive mode"
                    )
                )
                sys.exit(0)
            else:
                sys.exit(1)


@dataclass
class BodyError:
    """
    Encapsulates the errors related to commit message body validation.

    :attributes:
        is_conventional (List[str]): List of commit description lines that start with a conventional commit.
        descr_is_too_short (List[str]): List of commit description lines that are too short.
        num_empty_line (int): Number of empty lines in the commit message.
        invalid_body_length (bool): Indicates if the commit description is too long.
        duplicate_jama_refs (List[str]): List of duplicate JAMA items references.
        invalid_jama_refs (List[str]): List of invalid JAMA items references.
        impact_tag_is_missing (bool): Indicates if the commit description is missing an impact tag.
        impact_content_line_is_short (bool): Indicates if the impact tag line is too short.
    """

    is_conventional: List[str]
    descr_is_too_short: List[str]
    num_empty_line: int
    invalid_body_length: bool
    duplicate_jama_refs: List[str]
    invalid_jama_refs: List[str]
    impact_tag_is_missing: bool
    impact_content_line_is_short: bool



from typing import List

from lupin_grognard.core.cmd import Command, run_command
from lupin_grognard.core.config import COMMIT_DELIMITER
from lupin_grognard.core.tools.log_utils import die


class Git:
    def ensure_valid_git_repository(self) -> None:
        c = run_command(command="git rev-parse --is-inside-work-tree")
        if c.return_code != 0:
            die(msg="Not a git repository")

    def get_log(
        self, max_line_count: int = None, first_parent: bool = False, tag: str = None
    ) -> Command:
        format: str = (
            "hash>>%H%nauthor>>%aN%nauthor_mail>>%aE%nauthor_date>>%ct%ntitle>>%s%nbody>>%b<<body%n"
        )
        command_parts = ["git log"]

        if first_parent:
            command_parts.append("--first-parent")

        if tag:
            command_parts.append(f"{tag}..HEAD")

        command_parts.append(f'--format="{format}"{COMMIT_DELIMITER}')

        if max_line_count:
            command_parts.append(f"--max-count={max_line_count}")

        return run_command(command=" ".join(command_parts))

    def get_branch_name(self) -> str:
        return run_command(command="git branch --show-current").stdout

    def get_author_and_commiter_mail(self, hash: str) -> List[str]:
        c = run_command(command=f'git show --pretty=format:"%ae/%ce" -s {hash}')
        if c.return_code != 0:
            die(msg=f"Git error while getting author and committer mail: {c.stderr}")
        return c.stdout.split("/")

    def get_remote_origin_url(self) -> str:
        c = run_command(command="git config --get remote.origin.url")
        if c.return_code != 0:
            die(msg=f"Git error while getting remote origin url: {c.stderr}")
        gitlab_url = c.stdout
        if gitlab_url.startswith("https://gitlab.com/"):
            return gitlab_url[:-4] if gitlab_url.endswith(".git") else gitlab_url
        else:
            a = gitlab_url.find(":")
            if a != -1:
                gitlab_url = gitlab_url.replace(":", "/")
            gitlab_location = gitlab_url.find("@gitlab.com")
            gitlab_url = gitlab_url[gitlab_location + 1 :]
            gitlab_url = "https://" + gitlab_url
            return gitlab_url[:-4] if gitlab_url.endswith(".git") else gitlab_url

    def get_tags(self) -> List[List]:
        """Returns a list of tags with the following format:
        [
            ["tag_name", "tag_hash", "tag_date"],
            ["tag_name", "tag_hash", "tag_date"],
            ...
        ]
        """
        inner_delimiter = "---inner_delimiter---"
        dateformat = "%Y-%m-%d"
        formatter = (
            f'"%(refname:lstrip=2){inner_delimiter}'
            f"%(objectname){inner_delimiter}"
            f"%(creatordate:format:{dateformat}){inner_delimiter}"
            f'%(object)"'
        )
        c = run_command(command=f"git tag --format={formatter} --sort=-creatordate")

        if c.return_code != 0:
            die(msg=f"Git error while getting tags: {c.stderr}")
        if not c.stdout:
            return []

        tags_list = [line for line in c.stdout.splitlines()]
        return [tag.split(inner_delimiter)[:-1] for tag in tags_list]

    def get_parents(self, commit_hash: str) -> List[str]:
        c = run_command(command=f"git show --format=%P -s {commit_hash}")
        if c.return_code != 0:
            die(msg=f"Git error while getting parents of commit: {c.stderr}")
        return c.stdout.split(" ")

    def get_first_commit_date(self) -> str:
        c = run_command(
            'git log --reverse --format="%cd" --date="format-local:%d/%m/%y %I:%M %p"'
        )
        if c.return_code != 0:
            die(msg=f"Git error while getting first commit date: {c.stderr}")
        return c.stdout.split("\n")[0]

    def get_last_commit_date(self) -> str:
        c = run_command(
            'git log -1 --pretty="format:%cd" --date="format-local:%d/%m/%y %I:%M %p"'
        )
        if c.return_code != 0:
            die(msg=f"Git error while getting last commit date: {c.stderr}")
        return c.stdout

    def ensure_not_shallow_clone(self) -> None:
        """
        Check if the repository is a shallow clone.
        If the repository is a shallow clone, the function will print an error message and exit the program.
        """
        c = run_command(command="git rev-parse --is-shallow-repository")
        if c.stdout.strip() == "true":
            die(
                msg=(
                    "The repository is a shallow clone. If you are running from GitLab-CI, "
                    "configure your CI job with GIT_DEPTH=0."
                )
            )



import sys
from typing import List

from lupin_grognard.core.commit.commit import Commit, fill_authors_for_mr_commits
from lupin_grognard.core.commit.commit_error import ErrorCount
from lupin_grognard.core.commit.commit_validator import (
    CommitValidator,
    CommitCheckModes,
)
from lupin_grognard.core.tools.log_utils import die


def check_max_allowed_major_commits(
    commits: List[Commit], major_commit_limit: int
) -> bool:
    """Check if the number of major commits in `commits` exceeds `major_commit_limit`.

    Args:
        commits (List[Commit]): The list of commit object.
        major_commit_limit (int): The maximum number of major commits allowed.

    Returns:
        bool: True if the number of major commits is within the limit, else False.
    """
    if major_commit_limit == 0:  # --all option
        return True

    major_commit_count = 0
    for commit in commits:
        if commit.is_major_commit():
            major_commit_count += 1

    if major_commit_count > major_commit_limit:
        print(
            f"Error: found {major_commit_count} major commits to check in the "
            f"current branch while the maximum allowed number is {major_commit_limit}"
        )
        sys.exit(1)
    return True


def check_same_author_from_current_branch(commits: List[Commit]) -> None:
    """Check that all commits to be checked on the current branch come from the same author.

    Args:
        commits (List[Commit]): The list of commits to check.

    Returns:
        None
    """
    commits_author = set()

    for commit in commits:
        commits_author.update(commit.author_and_committer_mail)
        if len(commits_author) > 1:
            die(
                msg=(
                    "Found multiple authors for the commits in the current branch: please do not mix "
                    "commits from different authors in the same merge request."
                )
            )


def check_commit(
    commits: List[Commit],
    check_mode: CommitCheckModes,
    permissive_mode: bool,
    no_approvers: bool,
) -> None:
    """
    check_commit performs validation checks on each commit.
    If merge_option is set to 0, the function checks that merge commits
    have approvers.
    If merge_option is 1, the function only validates the title for a merge,
    the title and the body of the commit if it is a simple commit.
    The function also calls the error_report method of the ErrorCount
    class to output any errors found during validation.
    If any errors are found, it will call sys.exit(1)
    Args:
        commits (List): List of commits to check
        merge_option (int): 0 or 1
        permissive_mode (bool): If True, the function will not call sys.exit(1)
    """
    error_counter = ErrorCount()
    commits = [
        CommitValidator(
            commit=c,
            error_counter=error_counter,
            check_mode=check_mode,
            no_approvers=no_approvers,
        )
        for c in commits
    ]

    if check_mode == CommitCheckModes.CHECK_ALL_COMMITS:
        commits = fill_authors_for_mr_commits(commits=commits)

    for commit in commits:
        commit.perform_checks()

    error_counter.error_report(
        permissive_mode=permissive_mode,
    )



import logging
import os
import re
import sys
from typing import List

from lupin_grognard.core.commit.commit import Commit
from lupin_grognard.core.config import (
    COMMIT_DELIMITER,
    INITIAL_COMMIT,
    MAJOR_COMMIT_TYPES,
)
from lupin_grognard.core.git import Git


def read_file(file: str) -> str:
    with open(f"{file}", "r", encoding="utf-8") as file:
        data = file.read()
        return data


def write_file(file: str, content: str):
    with open(f"{file}", "w", encoding="utf-8") as outfile:
        outfile.write(content)


def get_version():
    """get version from setup.cfg file and
    update __version__ in lupin_grognard.__init__.py
    """
    setup_cfg = read_file("setup.cfg")
    _version = re.search(
        r"(^version = )(\d{1,2}\.\d{1,2}\.\d{1,2})(\.[a-z]{1,})?(\d{1,2})?",
        setup_cfg,
        re.MULTILINE,
    )
    version = ""
    for group in _version.group(2, 3, 4):
        if group is not None:
            version = version + str(group)
    content = f'__version__ = "{version}"\n'
    write_file(file="lupin_grognard/__init__.py", content=content)
    return version


def get_current_branch_name() -> str:
    branch_name = Git().get_branch_name()
    # branch name can be messing if running in CI
    if not branch_name:
        branch_name = os.getenv("CI_COMMIT_BRANCH")
    if not branch_name:
        branch_name = os.getenv("CI_MERGE_REQUEST_SOURCE_BRANCH_NAME")
    if not branch_name:
        return ""
    return branch_name


def display_supported_commit_types() -> None:
    commit_type = [
        "build",
        "bump",
        "ci",
        "deps(add|change|remove)",
        "docs",
        "enabler",
        "feat(add|change|remove)",
        "fixbug",
        "fixdefect(JAMA-xxx)",
        "refactor",
        "test",
    ]
    print("Supported commit types: " + ", ".join(commit_type))
    print(
        "Only one commit of a major type is allowed per merge branch: "
        f'{", ".join(MAJOR_COMMIT_TYPES[:-1])}, or {MAJOR_COMMIT_TYPES[-1]}.'
    )


def display_current_branch_name(current_branch_name: str) -> None:
    if current_branch_name:
        print(f"Current branch name is '{current_branch_name}'")
    else:
        print("Current branch name is not available")


def display_number_of_commits_to_check(commits: List[Commit]):
    number_commits_to_check = len(commits)
    if number_commits_to_check == 0:
        print("0 commit to check")
        sys.exit(0)
    elif number_commits_to_check == 1:
        print(f"Found {number_commits_to_check} commit to check:")
    else:
        print(f"Found {number_commits_to_check} commits to check:")


def generate_commit_list(commits_string: str) -> List[Commit]:
    """Geneartes the list of commits from Git().get_log().stdout value"""
    return [
        Commit(commit)
        for commit in commits_string.split(COMMIT_DELIMITER)
        if len(commit) > 0
    ]


def generate_commits_range_to_check(
    branch_list: List[str],
    commits: List[Commit],
    ci_mr_target_branch_name: str = "",
    initial_commits: List[str] = INITIAL_COMMIT,
) -> List:
    """
    generates a list of message ranges starting with INITIAL_COMMIT
    or the last merge into a branch contained in the branch list

    If ci_mr_target_branch_name is a main one, the function returns the commit range
    starting from the second merge found.

    If ci_mr_target_branch_name is a dev branches, the function returns the commit range
    starting with the second merge found and ending with the first commit after the first merge.
    """
    merge_count = 0
    merge_result_pipeline_main_branches = False
    merge_result_pipeline_dev_branches = False

    for index, commit in enumerate(commits):
        if commit.is_gitlab_merge_commit() and index == 0:
            if ci_mr_target_branch_name in ["main", "master"]:
                merge_result_pipeline_main_branches = True
            elif ci_mr_target_branch_name in ["dev", "develop", "development"]:
                merge_result_pipeline_dev_branches = True

        if merge_result_pipeline_main_branches:
            if commit.is_gitlab_merge_commit():
                merge_count += 1
                if merge_count == 2:
                    return commits[:index]  # with first merge commit
            elif commit.title in initial_commits:
                return commits[:index]
        elif merge_result_pipeline_dev_branches:
            if commit.is_gitlab_merge_commit():
                merge_count += 1
                if merge_count == 2:
                    return commits[1:index]  # skip first merge commit
            elif commit.title in initial_commits:
                return commits[1:index]
        else:
            if commit.is_gitlab_merge_commit():
                for branch in branch_list:
                    if commit.title.endswith(f"into '{branch}'"):
                        return commits[:index]
            elif commit.title in initial_commits:
                return commits[:index]
    return list()


def configure_logging():
    logging.basicConfig(
        format="%(asctime)s %(levelname)s: %(message)s", level=logging.INFO
    )



import os
from typing import List, Union

from lupin_grognard.core.tools.ros2.interfaces import Action, Msg, Srv
from lupin_grognard.core.tools.log_utils import die


def find_ros_packages(path: str) -> List[str]:
    """
    Retrieves the names of ROS packages present in the directory `path` and its subdirectories.

    :param path: Absolute path of the directory to explore.
    :type path: str
    :return: List of ROS package names, corresponding to directories containing a "package.xml" file.
    :rtype: List[str]
    """
    package_xml_files = []
    for root, dirs, files in os.walk(path):
        for file in files:
            if file == "package.xml":
                package_xml_files.append(root)

    if len(package_xml_files) == 0:
        raise die(msg=f"No ROS package found in {path}")
    return package_xml_files


class PackageRos2:
    def __init__(self, package_path: str):
        self.package_path = package_path
        self.name = os.path.basename(package_path)
        self.description = self._get_package_description()

    def _get_package_description(self) -> str:
        package_xml = os.path.join(self.package_path, "package.xml")
        if os.path.isfile(package_xml):
            with open(package_xml, "r", encoding="utf-8") as f:
                for line in f:
                    if "<description>" in line:
                        description = line.split("<description>")[1].split(
                            "</description>"
                        )[0]
                        if description.endswith("."):
                            description = description.rstrip(".")
                        return description
        return ""

    def _find_files(self, file_type: str) -> List[Union[Msg, Srv, Action]]:
        """
        Finds all files of a given type in the package directory.

        :param file_type: Type of files to search for, either "msg", "srv", or "action".
        :type file_type: str
        :return: A list of parsed message/service/action objects corresponding to the files found.
        :rtype: List[Union[Msg, Srv, Action]]
        """
        files = []
        match_dict = {
            "msg": Msg,
            "srv": Srv,
            "action": Action,
        }
        file_dir = os.path.join(self.package_path, file_type)
        if os.path.isdir(file_dir):
            for file in os.listdir(file_dir):
                if file.endswith(f".{file_type}"):
                    files.append(os.path.join(file_dir, file))
        return [match_dict[file_type](self.name, file) for file in files]



import logging
import sys


def die(msg: str) -> None:
    logging.error(msg)
    sys.exit(1)


def warn(msg: str) -> None:
    logging.warning(msg)


def info(msg: str) -> None:
    logging.info(msg)
